// Copyright 2012 Masanori Morise. All Rights Reserved.
// Author: morise [at] fc.ritsumei.ac.jp (Masanori Morise)
//
// Test program for WORLD 0.1.2 (Aug. 19)

// Platinum: 2010/Sept28test/Nov23.m (Private memo)
// test.exe input.wav outout.wav f0 spec
// input.wav  : Input file
// output.wav  : Output file
// f0      : F0 scaling (a positive number)
// spec      : Formant shift (a positive number)
// flag      : 0: Excitation signal (0.0.4), 1: Aperiodicity (0.0.3)

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#if (defined (__WIN32__) || defined (_WIN32)) && !defined (__MINGW32__)
#include <windows.h>
#include <conio.h>
#pragma comment(lib, "winmm.lib")
#endif
#if (defined (__linux__) || defined(__CYGWIN__))
// Linux porting section: include necessary header files under Linux system
#include <sys/time.h>
#include <termios.h>
#include <unistd.h>
#include <stdint.h>
#ifndef ECHOPRT
#define ECHOPRT ECHOE
#endif
#endif

#include "./dio.h"
#include "./star.h"
#include "./platinum.h"
#include "./tandem_ap.h"
#include "./synthesis.h"
#include "./synthesis_ap.h"
#include "./matlabfunctions.h"

#pragma warning(disable : 4996)

// Frame shift [msec]
#define FRAMEPERIOD 2.0

#if (defined (__linux__) || defined(__CYGWIN__))
// Linux porting section: implement timeGetTime() by gettimeofday(),
//                        implement getch() by termios
#ifndef DWORD
#define DWORD uint32_t
#endif
DWORD timeGetTime() {
  struct timeval tv;
  gettimeofday(&tv, NULL);
  DWORD ret = static_cast<DWORD>(tv.tv_usec / 1000 + tv.tv_sec * 1000);
  return ret;
}
int getch() {
  int ch;
  struct termios oldt, newt;

  tcgetattr(STDIN_FILENO, &oldt);
  memcpy(&newt, &oldt, sizeof(newt));
  newt.c_lflag &=
    ~(ECHO | ICANON | ECHOE | ECHOK | ECHONL | ECHOPRT | ECHOKE | ICRNL);
  tcsetattr(STDIN_FILENO, TCSANOW, &newt);
  ch = getchar();
  tcsetattr(STDIN_FILENO, TCSANOW, &oldt);

  return ch;
}
#endif

//-----------------------------------------------------------------------------
// Test program.
// test.exe input.wav outout.wav f0 spec flag
// input.wav  : argv[1] Input file
// output.wav : argv[2] Output file
// f0         : argv[3] F0 scaling (a positive number)
// spec       : argv[4] Formant shift (a positive number)
// flag       : argv[5] Excitation signal or Aperiodicity(0 or 1)
//-----------------------------------------------------------------------------
int main(int argc, char *argv[]) {
  if(argc != 2 && argc != 3 && argc != 4 && argc != 5 && argc != 6) 
  {
    printf("error: ̐słD\n");
    return 0;
  }
  int fs, nbit, x_length;
  double *x = wavread(argv[1], &fs, &nbit, &x_length);

  if (x == NULL) {
    printf("error: File not found.\n");
    getch();
    return 0;
  }

  printf("File information\n");
  printf("Sampling : %d Hz %d Bit\n", fs, nbit);
  printf("Length %d [sample]\n", x_length);
  printf("Length %f [sec]\n", static_cast<double>(x_length) / fs);

  // The number of samples is calculated in advance.
  int f0_length = GetSamplesForDIO(fs, x_length, FRAMEPERIOD);
  double *f0 = new double[f0_length];
  double *time_axis = new double[f0_length];
  double *y = new double[x_length];

  // F0 estimation
  DioOption option;  // 0.1.1
  InitializeDioOption(&option);  // Initialize the option
  // modification of the option
  option.frame_period = FRAMEPERIOD;
  option.speed = 2;
  option.f0_floor = 80.0;
  option.f0_ceil = 640.0;
//  option.f0_ceil = 800.0;
  option.channels_in_octave = 4.0;

  printf("\nAnalysis\n");
  DWORD elapsedTime = timeGetTime();
//  Dio(x, signalLen, fs, FRAMEPERIOD, t, f0); // version 0.1.0
  Dio(x, x_length, fs, option, time_axis, f0);
  printf("DIO: %d [msec]\n", timeGetTime() - elapsedTime);

  FILE *tmp_fp = fopen("f0info.dat", "w");
  for (int i = 0; i < f0_length; i++)
    fprintf(tmp_fp, "%f\n", f0[i]);
  fclose(tmp_fp);

  int fft_size = GetFFTSizeForStar(fs);

  double **spectrogram = new double *[f0_length];
  double **residual_spectrogram = new double *[f0_length];
  // for aperiodicity
  int number_of_bands = GetNumberOfBands(fs);
  double **aperiodicity = new double *[f0_length];
  for (int i = 0; i < f0_length; ++i) {
    spectrogram[i] = new double[fft_size / 2 + 1];
    residual_spectrogram[i] = new double[fft_size + 1];  // miss??
    aperiodicity[i] = new double[number_of_bands];
  }

  // Spectral envelope estimation
  elapsedTime = timeGetTime();
  Star(x, x_length, fs, time_axis, f0, f0_length, spectrogram);
  printf("STAR: %d [msec]\n", timeGetTime() - elapsedTime);

  bool tandem_flag = false;
  if (argc == 6) if (argv[5][0] == '1') tandem_flag = true;

  double target_f0;
  if (tandem_flag == false) {
    // Excitation signal.
    elapsedTime = timeGetTime();
    Platinum(x, x_length, fs, time_axis, f0, f0_length,
        spectrogram, fft_size, residual_spectrogram);
    printf("PLATINUM: %d [msec]\n", timeGetTime() - elapsedTime);
  } else {
    // Aperiodicity.
    elapsedTime = timeGetTime();
    target_f0 = AperiodicityRatio(x, x_length, fs, f0, f0_length,
        FRAMEPERIOD, aperiodicity);
    printf("TANDEM_AP: %d [msec]\n", timeGetTime() - elapsedTime);
  }

  // Note that F0 must not be changed until all parameters are estimated.

  // F0 scaling
  if (argc >= 4) {
    double shift = atof(argv[3]);
    for (int i = 0; i < f0_length; ++i) f0[i] *= shift;
  }
  // Spectral stretching
  if (argc >= 5) {
    double ratio = atof(argv[4]);
    double *freq_axis1 = new double[fft_size];
    double *freq_axis2 = new double[fft_size];
    double *spectrum1 = new double[fft_size];
    double *spectrum2 = new double[fft_size];

    for (int i = 0; i <= fft_size / 2; ++i) {
      freq_axis1[i] = ratio * i / fft_size * fs;
      freq_axis2[i] = static_cast<double>(i) / fft_size * fs;
    }
    for (int i = 0; i < f0_length; ++i) {
      for (int j = 0; j <= fft_size / 2; ++j)
        spectrum1[j] = log(spectrogram[i][j]);
      interp1(freq_axis1, spectrum1, fft_size / 2 + 1, freq_axis2,
        fft_size / 2 + 1, spectrum2);
      for (int j = 0; j <= fft_size / 2; ++j)
        spectrogram[i][j] = exp(spectrum2[j]);
      if (ratio < 1.0) {
        for (int j = static_cast<int>(fft_size / 2.0 * ratio);
            j <= fft_size / 2; ++j)
          spectrogram[i][j] =
          spectrogram[i][static_cast<int>(fft_size / 2.0 * ratio) - 1];
      }
    }
    delete[] spectrum1;
    delete[] spectrum2;
    delete[] freq_axis1;
    delete[] freq_axis2;
  }

  if (tandem_flag == false) {
    // Synthesis
    printf("\nSynthesis\n");
    elapsedTime = timeGetTime();
    Synthesis(f0, f0_length, spectrogram, residual_spectrogram,
        fft_size, FRAMEPERIOD, fs, x_length, y);
    printf("WORLD: %d [msec]\n", timeGetTime() - elapsedTime);
  } else {
    // Synthesis by the aperiodicity
    printf("\nSynthesis\n");
    elapsedTime = timeGetTime();
    SynthesisFromAperiodicity(f0, f0_length, spectrogram, fft_size,
        aperiodicity, number_of_bands, target_f0, FRAMEPERIOD, fs,
        x_length, y);
    printf("WORLD: %d [msec]\n", timeGetTime() - elapsedTime);
  }

  // Output
  wavwrite(y, x_length, fs, 16, argv[2]);

  printf("complete.\n");

  delete[] x;
  delete[] time_axis;
  delete[] f0;
  delete[] y;
  for (int i = 0; i < f0_length; ++i) {
    delete[] spectrogram[i];
    delete[] residual_spectrogram[i];
    delete[] aperiodicity[i];
  }
  delete[] spectrogram;
  delete[] residual_spectrogram;
  delete[] aperiodicity;

//  getch();
  return 0;
}
